(* Author: Andreas Lochbihler, ETH Zurich *)

theory Monom_Order
imports Nat_Mapping Compare
begin

instantiation nat_mapping :: ("{linorder_compare, zero}") linpreorder_compare 
begin
definition compare_nat_mapping :: 
  "'a nat_mapping comparator" where "compare_nat_mapping = comparator_of_le"
instance by(rule linpreorder_compare_classI)(simp add: compare_nat_mapping_def)
end

instance nat_mapping :: ("{linorder_compare, zero}") linorder_compare ..

(* TODO: direct implementation of compare on nat_mapping *)




type_synonym monom = "nat nat_mapping"

text \<open>
  The following typedef allows all linear orders as monomial orders.
  This should be refined to the proper notion of monomial order.
\<close>

typedef monom_order = "{compare :: monom comparator. comparator_eq compare}"
  morphisms monom_compare Abs_monom_order
by(auto intro: comparator_eq_on_comparator_of_le)

setup_lifting type_definition_monom_order

lemma comparator_eq_on_monom_compare [simp, intro!]:
  "comparator_eq_on (monom_compare mo) C"
proof -
  interpret comparator_eq_on "monom_compare mo" UNIV 
    using monom_compare[of mo] by simp
  show ?thesis by(rule subset_comparator_eq_on) simp
qed

lemma comparator_on_monom_compare [simp, intro!]:
  "comparator_on (monom_compare mo) C"
proof -
  interpret comparator_eq_on "monom_compare mo" UNIV 
    using monom_compare[of mo] by simp
  show ?thesis by(rule subset_comparator_on) simp
qed

lift_definition lex_mo :: monom_order is "comparator_of_le :: monom comparator"
by(rule comparator_eq_on_comparator_of_le)

lift_definition rlex_mo :: monom_order is "- comparator_of_le :: monom comparator"
by(rule comparator_eq_on.comparator_eq_on_dual)(rule comparator_eq_on_comparator_of_le)


code_datatype rlex_mo lex_mo

lemma [code]:
  shows monom_compare_lex: "monom_compare lex_mo = compare"
  and monom_compare_rlex: "monom_compare rlex_mo = - compare"
by(transfer, simp add: compare_nat_mapping_def)+

instantiation monom_order :: equal
begin
definition equal_monom_order :: 
  "monom_order \<Rightarrow> monom_order \<Rightarrow> bool" where "equal_monom_order = op ="
instance by(intro_classes)(simp add: equal_monom_order_def)
end

lemma equal_monom_order_code [code]:
  "HOL.equal lex_mo  lex_mo  = True"
  "HOL.equal lex_mo  rlex_mo = False"
  "HOL.equal rlex_mo lex_mo  = False"
  "HOL.equal rlex_mo rlex_mo = True"
apply(simp_all add: equal_eq)
sorry


typedef ('a, 'b) ignore_arg = "{f :: 'a \<Rightarrow> 'b. \<forall>x y. f x = f y}" 
  morphisms apply_ignore Abs_ignore_arg
by auto

setup_lifting type_definition_ignore_arg

lemma apply_ignore_ignore: "apply_ignore f x = apply_ignore f y"
by transfer simp


datatype impl = IMPL

lemma impl_eq [no_atp]: "x = IMPL" by(cases x)(simp)

text \<open>
  Simplification procedure for @{thm [source] impl_eq}.  Cannot use
  this rule directly --- it loops!
\<close>

simproc_setup impl_eq ("x::impl") = \<open>fn _ => fn _ => fn ct =>
  case Thm.term_of ct of Const (@{const_name IMPL}, _) => NONE | _ => SOME (mk_meta_eq @{thm impl_eq})
\<close>

lemma equal_impl_code [code]:
  "HOL.equal (x :: impl) y = True"
by(simp add: equal_eq)

definition RLEX :: impl where [simp, code del]: "RLEX = IMPL"
definition LEX :: impl where [simp, code del]: "LEX = IMPL"

code_datatype RLEX LEX

definition implT :: "monom_order \<Rightarrow> impl"
where [simp, code del]: "implT _ = IMPL"

hide_const (open) implT

adhoc_overloading implT Monom_Order.implT

lemma implT_code [code]:
  "implT lex_mo = LEX"
  "implT rlex_mo = RLEX"
by simp_all

lift_definition switch_monom_order :: "(monom_order, 'a) ignore_arg \<Rightarrow> impl \<Rightarrow> 'a"
is "\<lambda>f :: monom_order \<Rightarrow> 'a. \<lambda>_. f undefined" .

lemma switch_monom_order_code [code]:
  "switch_monom_order f LEX = apply_ignore f lex_mo"
  "switch_monom_order f RLEX = apply_ignore f rlex_mo"
by(simp_all add: switch_monom_order_def apply_ignore_ignore)

lemma switch_monom_order_apply [simp]:
  "switch_monom_order f mo = apply_ignore f undefined"
by transfer simp

hide_const (open) IMPL
hide_type (open) impl

end