(* Author: Florian Haftmann, TU Muenchen *)

section \<open>Some material relating polynomial mappings to ordered association lists\<close>

theory OAList_Poly_Mapping
imports OAList Poly_Mapping
begin

context
  fixes cmp :: "'a comparator"
  and dflt :: "'b"
begin

definition dest_fun_by :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a \<times> 'b) list"
where "dest_fun_by f = map (\<lambda>k. (k, f k)) (c.sorted_list_of_set_by cmp {k. f k \<noteq> dflt})"

lemma dest_fun_by_const [simp]: "dest_fun_by (\<lambda>_. dflt) = []"
by(simp add: dest_fun_by_def)

context
  fixes f :: "'a \<Rightarrow> 'b"
  assumes fin: "finite {k. f k \<noteq> dflt}"
  and comparator_eq: "comparator_eq cmp"
begin

lemma dsorted_by_dest_fun_by:
  "c.dsorted_by (compare_vimage fst cmp) (dest_fun_by f)"
using fin comparator_eq
by(simp add: dsorted_by_vimage dest_fun_by_def o_def comparator_eq_on.dsorted_by_sorted_list_of_set_by)

lemma set_dest_fun_by: "set (dest_fun_by f) = {(k, f k)|k. f k \<noteq> dflt}"
using fin comparator_eq
by(auto simp add: dest_fun_by_def comparator_eq_on.set_sorted_list_of_set_by)

interpretation c: comparator_eq_on cmp UNIV by(fact comparator_eq)

lemma dest_fun_by_fun_upd:
  "dest_fun_by (f(x := y)) =
   (if y = dflt then c.remove_by (compare_vimage fst cmp) (x, y) (dest_fun_by f)
    else c.update_by cmp x y (dest_fun_by f))"
proof(cases "y = dflt")
  case True
  have fin1: "finite {k. k \<noteq> x \<and> f k \<noteq> dflt}" using fin by(rule finite_subset[rotated]) auto
  have [simp]: "inj (\<lambda>k. (k, f k))" by(auto intro: injI)
  have "dest_fun_by (f(x := y)) = map (\<lambda>k. (k, (f(x := y)) k)) (c.sorted_list_of_set_by cmp {k. k \<noteq> x \<and> f k \<noteq> dflt})"
    using True by(simp add: dest_fun_by_def cong: conj_cong)
  also have "\<dots> = map (\<lambda>k. (k, f k)) (c.sorted_list_of_set_by cmp {k. k \<noteq> x \<and> f k \<noteq> dflt})"
    by(rule map_cong[OF refl])(simp_all add: c.set_sorted_list_of_set_by fin1)
  also have "c.sorted_list_of_set_by cmp {k. k \<noteq> x \<and> f k \<noteq> dflt} = c.remove_by cmp x (c.sorted_list_of_set_by cmp {k. f k \<noteq> dflt})"
    by(rule c.dsorted_by_unique)(auto simp add: c.set_sorted_list_of_set_by c.dsorted_by_sorted_list_of_set_by fin fin1 c.dsorted_by_remove_by c.set_remove_by)
  also have "map (\<lambda>k. (k, f k)) \<dots> = c.remove_by (compare_vimage (inv (\<lambda>k. (k, f k))) cmp) (x, f x) (dest_fun_by f)"
    unfolding dest_fun_by_def by(simp add: map_remove_by)
  also have "\<dots> = c.remove_by (compare_vimage fst cmp) (x, dflt) (dest_fun_by f)"
    by(rule remove_by_cong[OF refl])(auto simp add: dest_fun_by_def c.set_sorted_list_of_set_by fin compare_vimage_def intro!: arg_cong2[where f=cmp] inv_f_f)
  finally show ?thesis using True by(simp add: dest_fun_by_def)
next
  case False
  have "{k. k \<noteq> x \<longrightarrow> f k \<noteq> dflt} = insert x {k. f k \<noteq> dflt}" by auto
  hence "dest_fun_by (f(x := y)) = map (\<lambda>k. (k, (f(x := y)) k)) (c.sorted_list_of_set_by cmp (insert x {k. f k \<noteq> dflt}))"
    using False by(simp add: dest_fun_by_def)
  also have "\<dots> = map (\<lambda>k. (k, (f(x := y)) k)) (c.insort_by cmp x (c.sorted_list_of_set_by cmp {k. f k \<noteq> dflt}))"
    by(simp add: c.sorted_list_of_set_by_insert fin)
  also have "c.insort_by cmp x (c.sorted_list_of_set_by cmp {k. f k \<noteq> dflt}) = map fst (c.update_by cmp x y (dest_fun_by f))"
    by(rule c.dsorted_by_unique)(auto simp add: c.dsorted_by_insort_by c.dsorted_by_sorted_list_of_set_by fin dsorted_by_vimage[symmetric] dsorted_by_update_by[OF c.comparator_on] dsorted_by_dest_fun_by c.set_insort_by c.set_sorted_list_of_set_by set_dest_fun_by set_update_by[OF comparator_eq] intro: rev_image_eqI)
  also have "map (\<lambda>k. (k, (f(x := y)) k)) \<dots> = c.update_by cmp x y (dest_fun_by f)"
    by(auto intro!: map_idI simp add: set_update_by[OF comparator_eq] dsorted_by_dest_fun_by set_dest_fun_by)
  finally show ?thesis using False by(simp add: dest_fun_by_def)
qed

end

lemma dest_fun_by_lookup_default:
  assumes sorted: "c.dsorted_by (compare_vimage fst cmp) xs"
  and dflt: "dflt \<notin> snd ` set xs"
  and cmp: "comparator_eq cmp"
  shows "dest_fun_by (c.lookup_default cmp dflt xs) = xs"
proof -
  interpret c: comparator_eq_on cmp UNIV by(fact cmp)
  from dflt have "{k. c.lookup_default cmp dflt xs k \<noteq> dflt} = fst ` set xs"
    by(auto simp add: dom_lookup_default[OF sorted cmp] intro: rev_image_eqI)
  moreover have "c.sorted_list_of_set_by cmp (fst ` set xs) = map fst xs"
    using sorted by(auto intro!: c.dsorted_by_unique intro: c.dsorted_by_sorted_list_of_set_by simp add: dsorted_by_vimage c.set_sorted_list_of_set_by)
  ultimately show ?thesis using sorted
    by(auto simp add: dest_fun_by_def dsorted_by_vimage c.lookup_default_alt_def c.lookup_conv_map_of o_def map_of_is_SomeI[OF c.dsorted_by_distinct] intro!: map_idI)
qed

lemma lookup_default_dest_fun_by:
  assumes fin: "finite {x. f x \<noteq> dflt}"
  and cmp: "comparator_eq cmp"
  shows "c.lookup_default cmp dflt (dest_fun_by f) = f"
proof
  interpret c: comparator_eq_on cmp UNIV by(fact cmp)

  fix k
  show "c.lookup_default cmp dflt (dest_fun_by f) k = f k"
    using fin
    by(simp add: dest_fun_by_def c.lookup_default_alt_def c.lookup_conv_map_of dsorted_by_vimage o_def c.dsorted_by_sorted_list_of_set_by c.set_sorted_list_of_set_by map_of_map_restrict restrict_map_def)
qed

end

lemma sorted_list_of_set_by_comparator_of_le:
  assumes "finite A"
  shows "c.sorted_list_of_set_by comparator_of_le A = sorted_list_of_set A"
proof(rule sorted_distinct_set_unique)
  from comparator_eq_on_comparator_of_le assms
  have "c.dsorted_by comparator_of_le (c.sorted_list_of_set_by comparator_of_le A)"
    by(rule comparator_eq_on.dsorted_by_sorted_list_of_set_by) auto
  thus "sorted (c.sorted_list_of_set_by comparator_of_le A)"
    "distinct (c.sorted_list_of_set_by comparator_of_le A)"
    by(simp_all add: dsorted_by_conv_sorted_distinct)
qed(simp_all add: assms dsorted_by_conv_sorted_distinct comparator_eq_on.set_sorted_list_of_set_by[where D=UNIV, OF comparator_eq_on_comparator_of_le])

lemma dest_fun_by_comparator_of_le_lookup:
  "dest_fun_by comparator_of_le 0 (Poly_Mapping.lookup f) = Poly_Mapping.items f"
  by (simp add: Poly_Mapping.items_def dest_fun_by_def sorted_list_of_set_by_comparator_of_le)

context
  fixes cmp :: "'a comparator"
  and cmp' :: "'a comparator"
  assumes comparator_eq: "comparator_eq cmp"
  and comparator_eq': "comparator_eq cmp'"
begin

interpretation c: comparator_eq_on cmp UNIV by(fact comparator_eq)
interpretation c': comparator_eq_on cmp' UNIV by(fact comparator_eq')

lemma sort_by_dest_fun_by:
  assumes fin: "finite {k. f k \<noteq> x}"
  shows "c.sort_by (compare_vimage fst cmp) (dest_fun_by cmp' x f) = dest_fun_by cmp x f"
proof -
  have "c.sort_by cmp (c.sorted_list_of_set_by cmp' {k. f k \<noteq> x}) = c.sorted_list_of_set_by cmp {k. f k \<noteq> x}"
    by(rule c.dsorted_by_unique)(simp_all add: c.dsorted_by_sorted_list_of_set_by fin c.set_sort_by c.set_sorted_list_of_set_by c'.set_sorted_list_of_set_by)
  thus ?thesis
    by(simp add: dest_fun_by_def sort_by_map compare_vimage_compose o_def)
qed

end

context
  fixes cmp :: "'a comparator"
  assumes comparator_eq: "comparator_eq cmp"
begin

interpretation c: comparator_eq_on cmp UNIV by(fact comparator_eq)

lemma oa_zip_with'_dest_fun_by:
  assumes fin_p: "finite {x. p x \<noteq> dflt}"
  and fin_q: "finite {x. q x \<noteq> dflt'}"
  and f: "\<And>x. p x \<noteq> dflt \<Longrightarrow> f x (p x) \<noteq> Some dflt''"
  and g: "\<And>x. q x \<noteq> dflt' \<Longrightarrow> g x (q x) \<noteq> Some dflt''"
  and h: "\<And>x. \<lbrakk> p x \<noteq> dflt; q x \<noteq> dflt' \<rbrakk> \<Longrightarrow> h x (p x) (q x) \<noteq> Some dflt''"
  shows
  "oa_zip_with' cmp f g h (dest_fun_by cmp dflt p) (dest_fun_by cmp dflt' q) =
   dest_fun_by cmp dflt''
     (\<lambda>x. if p x = dflt then if q x = dflt' then dflt'' else case g x (q x) of None \<Rightarrow> dflt'' | Some y \<Rightarrow> y
          else if q x = dflt' then case f x (p x) of None \<Rightarrow> dflt'' | Some y \<Rightarrow> y
          else case h x (p x) (q x) of None \<Rightarrow> dflt'' | Some y \<Rightarrow> y)"
  (is "_ = dest_fun_by _ _ ?pq")
proof -
  from fin_p fin_q have "finite ({x. p x \<noteq> dflt} \<union> {y. q y \<noteq> dflt'})" by simp
  hence fin_pq: "finite {x. ?pq x \<noteq> dflt''}" by(rule finite_subset[rotated]) auto
  show ?thesis using fin_pq
    apply(simp add: dest_fun_by_def oa_zip_with'_def oa_zip_with_map_value[OF c.comparator_eq_on] c.dsorted_by_sorted_list_of_set_by fin_p fin_q c.set_sorted_list_of_set_by filter_map o_def cong: if_cong split del: split_if)
    apply(rule map_cong)
     apply(rule c.dsorted_by_unique)
        apply(simp_all add: c.dsorted_by_sorted_list_of_set_by fin_p fin_q c.dsorted_by_merge_by c.dsorted_by_filter fin_pq c.set_merge_by c.set_sorted_list_of_set_by)
    apply(auto split: option.splits split_if_asm simp add: f g h)
    done
qed

end

definition of_oalist :: "'a comparator \<Rightarrow> ('a \<times> 'b :: zero) list \<Rightarrow> ('a \<Rightarrow>\<^sub>0 'b)"
where "of_oalist cmp xs = Abs_poly_mapping (c.lookup_default cmp 0 xs)"

definition to_oalist :: "'a comparator \<Rightarrow> ('a \<Rightarrow>\<^sub>0 'b) \<Rightarrow> ('a \<times> 'b :: zero) list"
where "to_oalist cmp p = dest_fun_by cmp 0 (Poly_Mapping.lookup p)"

context begin interpretation lifting_syntax .

lemma of_oalist_transfer [transfer_rule]:
  "(eq_onp comparator_eq ===> op = ===> pcr_poly_mapping op = op =) (\<lambda>cmp. c.lookup_default cmp 0) of_oalist"
by(simp add: cr_poly_mapping_def poly_mapping.pcr_cr_eq of_oalist_def Abs_poly_mapping_inverse finite_dom_lookup_default rel_fun_def eq_onp_def)

lemma to_oalist_transfer [transfer_rule]:
  "(eq_onp comparator_eq ===> pcr_poly_mapping op = op = ===> op =) (\<lambda>cmp. dest_fun_by cmp 0) to_oalist"
by(simp add: cr_poly_mapping_def rel_fun_def eq_onp_def to_oalist_def poly_mapping.pcr_cr_eq)

end

lemma eq_onp_comparator_eq: "comparator_eq cmp \<Longrightarrow> eq_onp comparator_eq cmp cmp"
by(simp add: eq_onp_def)

lemma eq_onp_comparator_eq_comparator_of_le [transfer_rule]:
  "eq_onp comparator_eq comparator_of_le (comparator_of_le :: _ :: linorder \<Rightarrow> _)"
by(simp add: eq_onp_def)

lemma of_oalist_Nil [simp]: "of_oalist cmp [] = 0"
by(simp add: of_oalist_def c.lookup_default_alt_def[abs_def] zero_poly_mapping_def)

context
  fixes cmp :: "'a comparator"
  assumes comparator_eq [simp]: "comparator_eq cmp"
begin

interpretation c: comparator_eq_on cmp UNIV by fact

lemma of_oalist_single:
  notes eq_onp_comparator_eq[OF comparator_eq, transfer_rule]
  shows "of_oalist cmp [(k, v)] = Poly_Mapping.single k v"
by(transfer)(auto simp add: fun_eq_iff when_def c.compare_refl dest!: c.compare_eq_EQUALD split: comparison.split)

lemma of_oalist_to_oalist [simp]:
  "of_oalist cmp (to_oalist cmp p) = p"
by(simp add: of_oalist_def to_oalist_def lookup_default_dest_fun_by Poly_Mapping.lookup_inverse)

lemma to_oalist_of_oalist [simp]:
  "\<lbrakk> c.dsorted_by (compare_vimage fst cmp) xs;
     0 \<notin> snd ` set xs \<rbrakk>
   \<Longrightarrow> to_oalist cmp (of_oalist cmp xs) = xs"
by(simp add: of_oalist_def to_oalist_def Abs_poly_mapping_inverse finite_dom_lookup_default dest_fun_by_lookup_default)

lemma dsorted_by_to_oalist: "c.dsorted_by (compare_vimage fst cmp) (to_oalist cmp f)"
by(simp add: to_oalist_def dsorted_by_dest_fun_by)

lemma set_to_oalist: "set (to_oalist cmp f) = (\<lambda>k. (k, Poly_Mapping.lookup f k)) ` Poly_Mapping.keys f"
  by (auto simp add: to_oalist_def set_dest_fun_by)

lemma zero_notin_to_oalist: "0 \<notin> snd ` set (to_oalist cmp f)"
by(auto simp add: to_oalist_def set_dest_fun_by)

lemma to_oalist_single: "to_oalist cmp (Poly_Mapping.single m c) = (if c = 0 then [] else [(m, c)])"
unfolding to_oalist_def
proof(transfer fixing: cmp)
  fix m and c :: 'b
  have "(\<lambda>k' :: 'a. if m = k' then c else 0) = ((\<lambda>_. 0)(m := c))"
    by(simp add: fun_eq_iff)
  thus "dest_fun_by cmp 0 (\<lambda>k'. c when m = k') = (if c = 0 then [] else [(m, c)])"
    by(simp add: dest_fun_by_fun_upd when_def)
qed

end

lemma to_oalist_comparator_of_le:
  "to_oalist comparator_of_le p = Poly_Mapping.items p"
  by (simp add: to_oalist_def Poly_Mapping.items_def dest_fun_by_comparator_of_le_lookup)

lemma lookup_of_oalist_comparator_of_le:
  "sorted (map fst xs) \<Longrightarrow> distinct (map fst xs) \<Longrightarrow>
    Poly_Mapping.lookup (of_oalist comparator_of_le xs) = c.lookup_default comparator_of_le 0 xs"
  unfolding of_oalist_def
  apply (rule Abs_poly_mapping_inverse)
  apply simp
  apply (rule finite_dom_lookup_default)
  apply (rule comparator_eq_on_comparator_of_le)
  done

lemma sorted_map_fst_to_oalist_comparator_of_le [simp]:
  "sorted (map fst (to_oalist comparator_of_le f))"
  by transfer (simp add: sorted_list_of_set_by_comparator_of_le dest_fun_by_def comp_def)

lemma distinct_map_fst_to_oalist_comparator_of_le [simp]:
  fixes f :: "'a::linorder \<Rightarrow>\<^sub>0 'b::zero"
  shows "distinct (map fst (to_oalist comparator_of_le f))"
  by transfer (simp add: sorted_list_of_set_by_comparator_of_le dest_fun_by_def comp_def distinct_map)

lemma zero_not_in_to_oalist_comparator_of_le [simp]:
  fixes f :: "'a::linorder \<Rightarrow>\<^sub>0 'b::zero"
  shows "(k, 0) \<notin> set (to_oalist comparator_of_le f)"
  by transfer (auto simp add: sorted_list_of_set_by_comparator_of_le dest_fun_by_def)

primrec smember :: "'a \<Rightarrow> 'a :: linorder list \<Rightarrow> bool"
where
  "smember x []     \<longleftrightarrow> False"
| "smember x (y#ys) \<longleftrightarrow> \<not> x < y \<and> (y < x \<longrightarrow> smember x ys)"

lemma smember_sorted [simp]: "sorted xs \<Longrightarrow> smember k xs \<longleftrightarrow> k \<in> set xs"
  by (induction xs) (auto simp add: sorted_Cons)

lemma sa_lookup_default_map_range:
  "c.lookup_default comparator_of_le y (map (\<lambda>k. (k, f k)) xs) k = (if smember k xs then f k else y)"
  by (induction xs) (auto simp add: comparator_of_le_def)

lemma sa_lookup_default_dest:
  "c.lookup_default comparator_of_le 0 (Poly_Mapping.items m) = Poly_Mapping.lookup m"
  unfolding Poly_Mapping.items_def by transfer (simp add: sa_lookup_default_map_range fun_eq_iff)

lemma comparator_on_base_lookup_default_comparator_of_le_not_default_iff:
  assumes "sorted (map fst xs)" and "distinct (map fst xs)" and "dflt \<notin> snd ` set xs"
  shows "comparator_on_base.lookup_default comparator_of_le dflt xs k \<noteq> dflt \<longleftrightarrow> k \<in> fst ` set xs"
  using assms apply (induct xs k rule: c.lookup_default.induct [of _ comparator_of_le])
  apply (simp_all split: comparison.split)
  apply auto
  unfolding comparator_of_le_def
  apply (auto simp add: not_le sorted_Cons image_def split: if_splits)
  done

lemma comparator_on_base_lookup_default_comparator_of_le_eq_default_iff:
  assumes "sorted (map fst xs)" and "distinct (map fst xs)" and "dflt \<notin> snd ` set xs"
  shows "comparator_on_base.lookup_default comparator_of_le dflt xs k = dflt \<longleftrightarrow> k \<notin> fst ` set xs"
  using assms comparator_on_base_lookup_default_comparator_of_le_not_default_iff by metis

lemma keys_of_oalist_comparator_of_le:
  "sorted (map fst xs) \<Longrightarrow> distinct (map fst xs) \<Longrightarrow> 0 \<notin> snd ` set xs \<Longrightarrow>
    Poly_Mapping.keys (of_oalist comparator_of_le xs) = fst ` set xs"
  by transfer (simp add: comparator_on_base_lookup_default_comparator_of_le_not_default_iff)

lemma of_oalist_comparator_of_le_Cons:
  assumes "xs \<noteq> [] \<Longrightarrow> k < Min (fst ` set xs)"
  shows "of_oalist comparator_of_le ((k, v) # xs) = Poly_Mapping.update k v (of_oalist comparator_of_le xs)"
using assms proof transfer
  fix xs :: "('a \<times> 'b) list" and k :: 'a and v :: 'b
  assume "(xs \<noteq> [] \<Longrightarrow> k < Min (fst ` set xs))"
  then show "comparator_on_base.lookup_default comparator_of_le 0 ((k, v) # xs) =
    (comparator_on_base.lookup_default comparator_of_le 0 xs)(k := v)"
  apply (induct xs k rule: comparator_on_base.lookup_default.induct [of _ comparator_of_le])
  apply (simp_all add: fun_eq_iff split: comparison.split)
  unfolding comparator_of_le_def
  apply (auto simp add: not_le sorted_Cons image_def split: if_splits)
  done
qed

lemma of_oalist_comparator_of_le_conv_the_value:
  assumes "sorted (map fst xs)" and "distinct (map fst xs)"
  shows "of_oalist comparator_of_le xs = Poly_Mapping.the_value xs"
proof (insert assms, rule sym, induct xs)
  case Nil then show ?case by simp
next
  case (Cons x xs)
  then obtain k v where x: "x = (k, v)" by (cases x) blast
  with Cons show ?case
    apply simp
    apply (subst of_oalist_comparator_of_le_Cons)
    apply (auto simp add: sorted_Cons intro: neq_le_trans)
  done
qed

end